searchData={"items":[{"type":"module","title":"LetMe","doc":"LetMe is library for defining and evaluating authorization rules and handling\nquery scopes and field redactions.\n\nThis module only defines auxiliary functions. The main functionality lies in\nthe `LetMe.Policy` module.","ref":"LetMe.html"},{"type":"function","title":"LetMe.filter_allowed_actions/4","doc":"Takes a list of rules and only returns the rules that would evaluate to `true`\nfor the given subject and object.\n\nThe object needs to be passed as a tuple, where the first element is the\nobject name, and the second element is the actual object, e.g.\n`{:article, %Article{}}`.\n\nIf you registered the schema module with `LetMe.Policy.object/3`, you can\npass the struct without tagging it with the object name, e.g. `%Article{}`.\n\nThis function is used internally by `c:LetMe.Policy.filter_allowed_actions/3`.","ref":"LetMe.html#filter_allowed_actions/4"},{"type":"function","title":"Example - LetMe.filter_allowed_actions/4","doc":"rules = MyApp.Policy.list_rules()\n\n    filter_allowed_actions(\n      rules,\n      %User{},\n      {:article, %Article{}},\n      MyApp.Policy\n    )","ref":"LetMe.html#filter_allowed_actions/4-example"},{"type":"function","title":"LetMe.filter_rules/2","doc":"Takes a list of rules and a list of filter options and returns a filtered\nlist of rules.\n\nThis function is used by `c:LetMe.Policy.list_rules/1`.","ref":"LetMe.html#filter_rules/2"},{"type":"function","title":"Filter options - LetMe.filter_rules/2","doc":"- `:object` - Matches an object exactly.\n- `:action` - Matches an action exactly.\n- `:allow` - Either a check name as an atom or a 2-tuple with the check name\n  and the options.\n- `:metadata` - Either a metadata name as an atom or a 2-tuple with the\n  metadata name and the metadata value.\n- `:deny` - Either a check name as an atom or a 2-tuple with the check name\n  and the options.\n\nIf an atom is passed as `allow` or `deny`, the atom is interpreted as a check\nname and all rules using the given check name are returned, regardless of\nwhether additional options are passed to the check. If a 2-tuple is passed,\nthe first tuple element must be the check name as an atom and the second\ntuple element must be the check options. In this case, all rules are returned\nthat use the given check with exactly the same options. In either case, rules\nthat have more checks in addition to the given one will also be returned.","ref":"LetMe.html#filter_rules/2-filter-options"},{"type":"function","title":"Examples - LetMe.filter_rules/2","doc":"iex> rules = [\n    ...>   %LetMe.Rule{action: :create, name: :article_create, object: :article},\n    ...>   %LetMe.Rule{action: :create, name: :category_create, object: :category}\n    ...> ]\n    iex> filter_rules(rules, object: :article)\n    [%LetMe.Rule{action: :create, name: :article_create, object: :article}]\n\n    iex> rules = [\n    ...>   %LetMe.Rule{\n    ...>     action: :create,\n    ...>     name: :article_create,\n    ...>     object: :article,\n    ...>     allow: [[role: :editor]]\n    ...>   },\n    ...>   %LetMe.Rule{\n    ...>     action: :update,\n    ...>     name: :article_update,\n    ...>     object: :article,\n    ...>     allow: [:own_resource, [role: :writer]]\n    ...>   }\n    ...> ]\n    iex> filter_rules(rules, allow: :own_resource)\n    [%LetMe.Rule{action: :update, name: :article_update, object: :article, allow: [:own_resource, [role: :writer]]}]\n    iex> match?([_, _], filter_rules(rules, allow: :role))\n    true\n    iex> filter_rules(rules, allow: {:role, :editor})\n    [%LetMe.Rule{action: :create, name: :article_create, object: :article, allow: [[role: :editor]]}]\n    iex> filter_rules(rules, allow: {:role, :writer})\n    [%LetMe.Rule{action: :update, name: :article_update, object: :article, allow: [:own_resource, [role: :writer]]}]","ref":"LetMe.html#filter_rules/2-examples"},{"type":"function","title":"LetMe.redact/3","doc":"Takes a struct or a list of structs and redacts fields depending on the\nsubject (user).\n\nUses the callback implementation for `c:LetMe.Schema.redacted_fields/3` in the\nstruct module.","ref":"LetMe.html#redact/3"},{"type":"function","title":"Options - LetMe.redact/3","doc":"- `:redact_value` - The value to be used for redacted fields. Defaults to\n  `:redacted`.\n\nAny additional options will be passed to `c:LetMe.Schema.redacted_fields/3`.","ref":"LetMe.html#redact/3-options"},{"type":"function","title":"Example - LetMe.redact/3","doc":"iex> article = %MyApp.Blog.Article{}\n    iex> user = %{id: 2, role: :user}\n    iex> redact(article, user)\n    %MyApp.Blog.Article{like_count: :redacted, title: \"Give us back our moon dust and cockroaches\", user_id: 1, view_count: :redacted}\n\n    iex> article = %MyApp.Blog.Article{}\n    iex> user = %{id: 2, role: :user}\n    iex> redact(article, user, redact_value: nil)\n    %MyApp.Blog.Article{like_count: nil, title: \"Give us back our moon dust and cockroaches\", user_id: 1, view_count: nil}\n\n    iex> articles = [\n    ...>   %MyApp.Blog.Article{},\n    ...>   %MyApp.Blog.Article{user_id: 2, title: \"Joey Chestnut is chomp champ\"}\n    ...> ]\n    iex> user = %{id: 2, role: :user}\n    iex> redact(articles, user)\n    [%MyApp.Blog.Article{like_count: :redacted, title: \"Give us back our moon dust and cockroaches\", user_id: 1, view_count: :redacted}, %MyApp.Blog.Article{like_count: 25, title: \"Joey Chestnut is chomp champ\", user_id: 2, view_count: :redacted}]","ref":"LetMe.html#redact/3-example"},{"type":"function","title":"LetMe.reject_redacted_fields/4","doc":"Removes redacted fields from a given list of fields.\n\nUses the `c:LetMe.Schema.redacted_fields/3` callback implementation of the\nstruct module to determine the fields to remove.","ref":"LetMe.html#reject_redacted_fields/4"},{"type":"function","title":"Examples - LetMe.reject_redacted_fields/4","doc":"iex> fields = [:like_count, :title, :user_id, :view_count]\n    iex> user = %{id: 1, role: :user}\n    iex> article = %MyApp.Blog.Article{}\n    iex> result = reject_redacted_fields(fields, article, user)\n    iex> Enum.sort(result)\n    [:like_count, :title, :user_id]\n\nThis can be useful as a safeguard to prevent accidentally casting fields the\nuser is not allowed to see and thereby nilifying or replacing them.\n\n    def update_changeset(%Article{} = article, attrs, %User{} = user) do\n      fields = LetMe.reject_redacted_fields(\n        [:title, :body, :internal_reference],\n        article,\n        user\n      )\n\n      article\n      |> cast(attrs, fields)\n      |> validate_required([:title, :body])\n    end\n\nIf a keyword list is given as a fourth argument, it is passed to\n`c:LetMe.Schema.redacted_fields/3`.","ref":"LetMe.html#reject_redacted_fields/4-examples"},{"type":"behaviour","title":"LetMe.Policy","doc":"This module defines a DSL for authorization rules and compiles these rules\nto authorization and introspection functions.","ref":"LetMe.Policy.html"},{"type":"behaviour","title":"Usage - LetMe.Policy","doc":"defmodule MyApp.Policy do\n      use LetMe.Policy\n\n      object :article do\n        # Creating articles is allowed if the user role is `editor` or `writer`.\n        action :create do\n          allow role: :editor\n          allow role: :writer\n        end\n\n        # Viewing articles is always allowed, unless the user is banned.\n        action :read do\n          allow true\n          deny :banned\n        end\n\n        # Updating an article is allowed if (the user role is `editor`) OR\n        # (the user role is `writer` AND the article belongs to the user).\n        action :update do\n          allow role: :editor\n          allow [:own_resource, role: :writer]\n        end\n\n        # Deleting an article is allowed if the user is an editor.\n        action :delete do\n          allow role: :editor\n        end\n      end\n    end\n\n> #### `use LetMe.Policy` {: .info}\n>\n> When you `use LetMe.Policy`, the module will set `@behaviour LetMe.Policy`\n> and define all callback functions for that behaviour based on the macros\n> you use. It will also define an `action` type based your rules.\n\n#","ref":"LetMe.Policy.html#module-usage"},{"type":"behaviour","title":"Options - LetMe.Policy","doc":"These options can be passed when using this module:\n\n- `check_module` - The module where the check functions are defined. Defaults\n  to `__MODULE__.Checks`.\n- `error_reason` - The error reason used by the `c:authorize/4` callback.\n  Defaults to `:unauthorized`.\n- `error_message` - The error message used by the `c:authorize!/4`. Defaults\n  to \"unauthorized\".","ref":"LetMe.Policy.html#module-options"},{"type":"behaviour","title":"Check module - LetMe.Policy","doc":"The checks passed to `allow/1` and `deny/1` reference the names of functions\nin the check module.\n\nBy default, LetMe tries to find the functions in `__MODULE__.Checks` (in the\nexample, this would be `MyApp.Policy.Checks`). However, you can override the\ndefault check module:\n\n    use LetMe.Policy, check_module: MyApp.AuthChecks\n\nEach check function has to take the subject (user), the object, and optionally\nan additional argument, and must return a boolean value.\n\nFor example, this check determines whether a user is banned:\n\n    def banned(%User{banned: true}, _), do: true\n    def banned(%User{}, _), do: false\n\nThis check determines whether the user has the given role:\n\n    def role(%User{role: role}, _, role), do: true\n    def role(_, _, _), do: false\n\nAnd this check determines whether the object belongs to the user:\n\n    def own_resource(%User{id: user_id}, %{user_id: user_id}), do: true\n    def own_resource(_, _), do: false\n\nLetMe does not make any assumptions about your access control model, as long\nas you can map your rules to subject, object and action. You can use the three\nrules above with the `allow/1` and `deny/1` macros.\n\n    allow role: :admin\n    allow :own_resource\n    deny :banned","ref":"LetMe.Policy.html#module-check-module"},{"type":"behaviour","title":"Combining checks - LetMe.Policy","doc":"Rules evaluate to `false` by default. These rules will always be `false`\nbecause they don't have any `allow` clauses:\n\n    action :create do\n    end\n\n    action :update do\n      deny false\n    end\n\n\nTrying to evaluate a rule name that does not exist also evaluates to `false`.\n\nAs soon as one `deny` check evaluates to `true`, the whole rule will evaluate\nto `false`. This rule will always evaluate to `false`:\n\n    action :create do\n      allow true\n      deny true\n    end\n\nIf you pass a list of checks to either `allow/1` or `deny/1`, the checks\nare combined with a logical `AND`.\n\n    # false\n    action :create do\n      allow [true, false]\n    end\n\n    # true\n    action :create do\n      allow [true, true]\n    end\n\n    # true\n    action :create do\n      allow [true, true]\n      deny [true, false]\n    end\n\n    # false\n    action :create do\n      allow [true, true]\n      deny [true, true]\n    end\n\nOn the other hand, if either the `allow/1` or the `deny/1` macro is used\nmultiple times, the checks are combined with a logical `OR`.\n\n    # true\n    action :create do\n      allow true\n      allow false\n    end\n\n    # false\n    action :create do\n      allow [true, false]\n      allow false\n    end\n\n    # true\n    action :create do\n      allow [true, false]\n      allow true\n    end\n\n    # false\n    action :create do\n      allow [true, true]\n      allow true\n      deny false\n      deny true\n    end","ref":"LetMe.Policy.html#module-combining-checks"},{"type":"behaviour","title":"Pre-hooks - LetMe.Policy","doc":"You can use pre-hooks to process or gather additional data about the subject\nand/or object before running the checks. This can be useful if you need to\npreload associations or make external requests. Pre-hooks run once per\nauthorization request before running the checks. See the documentation for\n`pre_hooks/1`.","ref":"LetMe.Policy.html#module-pre-hooks"},{"type":"macro","title":"LetMe.Policy.action/2","doc":"Defines an action that needs to be authorized.\n\nWithin the do-block, you can use the `allow/1`, `deny/1` and `pre_hooks/1`\nmacros to define the checks to be run and the `desc/1` macro to add a\ndescription.\n\nThis macro must be used within the do-block of `object/2`.\n\nEach `action` block will be compiled to a rule. The rule name is an atom with\nthe format `{object}_{action}`.","ref":"LetMe.Policy.html#action/2"},{"type":"macro","title":"Example - LetMe.Policy.action/2","doc":"object :article do\n      action :create do\n        allow role: :editor\n        allow role: :writer\n      end\n\n      action :update do\n        allow role: :editor\n        allow [:own_resource, role: :writer]\n      end\n    end\n\nIf you have multiple actions with the same allow and deny rules, you can also\npass a list of action names as the first argument.\n\n    object :article do\n      action [:create, :update, :delete] do\n        allow role: :editor\n        allow role: :writer\n      end\n    end","ref":"LetMe.Policy.html#action/2-example"},{"type":"macro","title":"LetMe.Policy.allow/1","doc":"Defines the checks to be run to determine if an action is allowed.\n\nThe argument can be:\n\n- a function name as an atom\n- a tuple with the function name and an additional argument\n- a list of function names or function/argument tuples\n- `true` or `false` - Always allows or denies an action. Can be useful in\n  combination with the `deny/1` macro.\n\nThe function must be defined in the configured check module and take the\nsubject (current user), object as arguments, and if given, the additional\nargument.\n\nIf a list is given as an argument, the checks are combined with a logical\n`AND`.\n\nIf the `allow/1` macro is used multiple times within the same `action/2`\nblock, the checks of each macro call are combined with a logical `OR`.","ref":"LetMe.Policy.html#allow/1"},{"type":"macro","title":"Examples - LetMe.Policy.allow/1","doc":"Let's assume you defined the following checks:\n\n    defmodule MyApp.Policy.Checks do\n      def role(%User{role: role}, _, role), do: true\n      def role(_, _, _), do: false\n\n      def own_resource(%User{id: id}, %{user_id: id}, _), do: true\n      def own_resource(_, _, _), do: false\n    end\n\nThis would allow the `:article_update` action only if the current user has\nthe role `:admin`:\n\n    object :article do\n      action :update do\n        allow role: :admin\n      end\n    end\n\nThis is equivalent to:\n\n    object :article do\n      action :update do\n        allow {:role, :admin}\n      end\n    end\n\nThis would allow the `:article_update` action if the user has the role\n`:writer` _and_ the article belongs to the user:\n\n    object :article do\n      action :update do\n        allow [:own_resource, role: :writer]\n      end\n    end\n\nThis is equivalent to:\n\n    object :article do\n      action :update do\n        allow [:own_resource, {:role, :writer}]\n      end\n    end\n\nThis would allow the `:article_update` action if\n(the user has the role `:admin` _or_ (the user has the role `:writer` _and_\nthe article belongs to the user)):\n\n    object :article do\n      action :update do\n        allow role: :admin\n        allow [:own_resource, role: :writer]\n      end\n    end","ref":"LetMe.Policy.html#allow/1-examples"},{"type":"callback","title":"LetMe.Policy.authorize/4","doc":"Authorizes a request defined by the action, subject and object.","ref":"LetMe.Policy.html#c:authorize/4"},{"type":"callback","title":"Example - LetMe.Policy.authorize/4","doc":"Assume we defined this authorization rule:\n\n    object :article do\n      action :update do\n        allow :own_resource\n      end\n    end\n\nAnd the `:own_resource` check is defined as:\n\n    def own_resource(%{id: user_id}, %{user_id: user_id}), do: true\n    def own_resource(_, _), do: false\n\nThe rule name consists of the object and the action name, in this case\n`:article_create`. To authorize the action, we need to pass the rule name, the\nsubject (current user) and the object (the article to be updated).\n\n    iex> article = %{id: 80, user_id: 1}\n    iex> user_1 = %{id: 1}\n    iex> user_2 = %{id: 2}\n    iex> MyApp.Policy.authorize(:article_update, user_1, article)\n    :ok\n    iex> MyApp.Policy.authorize(:article_update, user_2, article)\n    {:error, :unauthorized}\n\nIf the checks don't require the object, it can be omitted.\n\n    object :user do\n      action :list do\n        allow {:role, :admin}\n        allow {:role, :client}\n      end\n    end\n\n    iex> user = %{id: 1, role: :admin}\n    iex> MyApp.Policy.authorize(:user_list, user)\n    :ok\n    iex> user = %{id: 2, role: :user}\n    iex> MyApp.Policy.authorize(:user_list, user)\n    {:error, :unauthorized}\n\nThe error reason can be customized by setting the `:error_reason` option when\nusing the module.\n\nThe last parameter is a set of arguments that can be defined dynamically\nwhich will be passed into any `pre_hook`s defined on the resource's policy.","ref":"LetMe.Policy.html#c:authorize/4-example"},{"type":"callback","title":"LetMe.Policy.authorize!/4","doc":"Same as `c:authorize/4`, but raises an error if unauthorized.","ref":"LetMe.Policy.html#c:authorize!/4"},{"type":"callback","title":"Example - LetMe.Policy.authorize!/4","doc":"With the same authorization rules as defined in the `c:authorize/4`\ndocumentation, we get this:\n\n    iex> article = %{id: 80, user_id: 1}\n    iex> user_1 = %{id: 1}\n    iex> user_2 = %{id: 2}\n    iex> MyApp.Policy.authorize!(:article_update, user_1, article)\n    :ok\n    iex> MyApp.Policy.authorize!(:article_update, user_2, article)\n    ** (LetMe.UnauthorizedError) unauthorized","ref":"LetMe.Policy.html#c:authorize!/4-example"},{"type":"callback","title":"LetMe.Policy.authorize?/4","doc":"Same as `c:authorize/4`, but returns a boolean.","ref":"LetMe.Policy.html#c:authorize?/4"},{"type":"callback","title":"Example - LetMe.Policy.authorize?/4","doc":"With the same authorization rules as defined in the `c:authorize/4`\ndocumentation, we get this:\n\n    iex> article = %{id: 80, user_id: 1}\n    iex> user_1 = %{id: 1}\n    iex> user_2 = %{id: 2}\n    iex> MyApp.Policy.authorize?(:article_update, user_1, article)\n    true\n    iex> MyApp.Policy.authorize?(:article_update, user_2, article)\n    false","ref":"LetMe.Policy.html#c:authorize?/4-example"},{"type":"macro","title":"LetMe.Policy.deny/1","doc":"Defines the checks to be run to determine if an action is denied.\n\nIf any of the checks evaluates to `true`, the `allow` checks are overridden\nand the authorization request is automatically denied.\n\nIf a list is given as an argument, the checks are combined with a logical\n`AND`.\n\nIf the `allow/1` macro is used multiple times within the same `action/2`\nblock, the checks of each macro call are combined with a logical `OR`.","ref":"LetMe.Policy.html#deny/1"},{"type":"macro","title":"Examples - LetMe.Policy.deny/1","doc":"Let's assume you defined the following checks:\n\n    defmodule MyApp.Policy.Checks do\n      def role(%User{role: role}, _, role), do: true\n      def role(_, _, _), do: false\n\n      def own_resource(%User{id: id}, %{user_id: id}, _), do: true\n      def own_resource(_, _, _), do: false\n\n      def same_user(%User{id: id}, %User{id: id}, _), do: true\n      def same_user(_, _, _), do: false\n    end\n\nThis would allow the `:user_delete` action by default, _unless_ the object is\nthe current user:\n\n    object :user do\n      action :delete do\n        allow true\n        deny :same_user\n      end\n    end\n\nThis would allow the `:article_update` action only if the current user has\nthe role `:admin`, _unless_ the object is the current user:\n\n    object :user do\n      action :delete do\n        allow role: :admin\n        deny :same_user\n      end\n    end\n\nThis would allow the `:user_delete` by default, _unless_ the object is the\ncurrent user _and_ the current user is an admin:\n\n    object :user do\n      action :delete do\n        allow true\n        deny [:same_user, role: :admin]\n      end\n    end\n\nThis would allow the `:user_delete` by default, _unless_ the object is the\ncurrent user _or_ the current user is a peasant:\n\n    object :user do\n      action :delete do\n        allow true\n        deny :same_user\n        deny role: :peasant\n      end\n    end","ref":"LetMe.Policy.html#deny/1-examples"},{"type":"macro","title":"LetMe.Policy.desc/1","doc":"Allows you to add a description to a rule.\n\nThe description can be accessed from the `LetMe.Rule` struct. You can use it\nto generate help texts or documentation.","ref":"LetMe.Policy.html#desc/1"},{"type":"macro","title":"Example - LetMe.Policy.desc/1","doc":"object :article do\n      action :create do\n        desc \"allows a user to create a new article\"\n        allow role: :writer\n      end\n    end","ref":"LetMe.Policy.html#desc/1-example"},{"type":"callback","title":"LetMe.Policy.fetch_rule/1","doc":"Returns the rule for the given name. Returns an `:ok` tuple or `:error`.\n\nThe rule name is an atom with the format `{object}_{action}`.","ref":"LetMe.Policy.html#c:fetch_rule/1"},{"type":"callback","title":"Example - LetMe.Policy.fetch_rule/1","doc":"iex> MyApp.Policy.fetch_rule(:article_create)\n    {:ok,\n     %LetMe.Rule{\n       action: :create,\n       allow: [[role: :admin], [role: :writer]],\n       deny: [],\n       name: :article_create,\n       object: :article,\n       pre_hooks: []\n     }}\n\n     iex> MyApp.Policy.fetch_rule(:cookie_eat)\n     :error","ref":"LetMe.Policy.html#c:fetch_rule/1-example"},{"type":"callback","title":"LetMe.Policy.fetch_rule!/1","doc":"Returns the rule with the given name. Raises if the rule is not found.\n\nThe rule name is an atom with the format `{object}_{action}`.","ref":"LetMe.Policy.html#c:fetch_rule!/1"},{"type":"callback","title":"Example - LetMe.Policy.fetch_rule!/1","doc":"iex> MyApp.Policy.fetch_rule!(:article_create)\n    %LetMe.Rule{\n      action: :create,\n      allow: [[role: :admin], [role: :writer]],\n      deny: [],\n      name: :article_create,\n      object: :article,\n      pre_hooks: []\n    }","ref":"LetMe.Policy.html#c:fetch_rule!/1-example"},{"type":"callback","title":"LetMe.Policy.filter_allowed_actions/3","doc":"Takes a list of rules and only returns the rules that would evaluate to `true`\nfor the given subject and object.","ref":"LetMe.Policy.html#c:filter_allowed_actions/3"},{"type":"callback","title":"Examples - LetMe.Policy.filter_allowed_actions/3","doc":"The object can be passed as a tuple, where the first element is the\nobject name, and the second element is the actual object, e.g.\n`{:article, %Article{}}`.\n\n    iex> rules = MyApp.Policy.list_rules()\n    iex> MyApp.Policy.filter_allowed_actions(\n    ...>   rules,\n    ...>   %{id: 2, role: nil},\n    ...>   {:article, %MyApp.Blog.Article{}}\n    ...> )\n    [\n      %LetMe.Rule{\n        action: :view,\n        allow: [true],\n        deny: [],\n        description: \"allows to view an article and the list of articles\",\n        name: :article_view,\n        object: :article,\n        pre_hooks: []\n      }\n    ]\n\nIf you registered the schema module with `LetMe.Policy.object/3`, you can\npass the struct without tagging it with the object name.\n\n    iex> rules = MyApp.Policy.list_rules()\n    iex> MyApp.Policy.filter_allowed_actions(\n    ...>   rules,\n    ...>   %{id: 2, role: nil},\n    ...>   %MyApp.Blog.Article{}\n    ...> )\n    [\n      %LetMe.Rule{\n        action: :view,\n        allow: [true],\n        deny: [],\n        description: \"allows to view an article and the list of articles\",\n        name: :article_view,\n        object: :article,\n        pre_hooks: []\n      }\n    ]","ref":"LetMe.Policy.html#c:filter_allowed_actions/3-examples"},{"type":"callback","title":"LetMe.Policy.get_object_name/1","doc":"Returns the object name for the given schema module or struct, if it was\nregistered using `object/3`.","ref":"LetMe.Policy.html#c:get_object_name/1"},{"type":"callback","title":"Examples - LetMe.Policy.get_object_name/1","doc":"iex> MyApp.Policy.get_object_name(MyApp.Blog.Article)\n    :article\n\n    iex> MyApp.Policy.get_object_name(%MyApp.Blog.Article{})\n    :article\n\n    iex> MyApp.Policy.get_object_name(MyApp.Blog.Tag)\n    nil","ref":"LetMe.Policy.html#c:get_object_name/1-examples"},{"type":"callback","title":"LetMe.Policy.get_rule/1","doc":"Returns the rule for the given rule name. Returns `nil` if the rule is\nnot found.\n\nThe rule name is an atom with the format `{object}_{action}`.","ref":"LetMe.Policy.html#c:get_rule/1"},{"type":"callback","title":"Example - LetMe.Policy.get_rule/1","doc":"iex> MyApp.Policy.get_rule(:article_create)\n    %LetMe.Rule{\n      action: :create,\n      allow: [[role: :admin], [role: :writer]],\n      deny: [],\n      name: :article_create,\n      object: :article,\n      pre_hooks: []\n    }\n\n    iex> MyApp.Policy.get_rule(:cookie_eat)\n    nil","ref":"LetMe.Policy.html#c:get_rule/1-example"},{"type":"callback","title":"LetMe.Policy.get_schema/1","doc":"Returns the schema module for the given object name, if it was registered\nusing `object/3`.","ref":"LetMe.Policy.html#c:get_schema/1"},{"type":"callback","title":"Examples - LetMe.Policy.get_schema/1","doc":"iex> MyApp.Policy.get_schema(:article)\n    MyApp.Blog.Article\n\n    iex> MyApp.Policy.get_schema(:user)\n    nil","ref":"LetMe.Policy.html#c:get_schema/1-examples"},{"type":"callback","title":"LetMe.Policy.list_rules/0","doc":"Returns all authorization rules as a list.","ref":"LetMe.Policy.html#c:list_rules/0"},{"type":"callback","title":"Example - LetMe.Policy.list_rules/0","doc":"iex> MyApp.PolicyShort.list_rules() |> Enum.sort()\n    [\n      %LetMe.Rule{\n        action: :create,\n        allow: [[role: :admin], [role: :writer]],\n        deny: [],\n        name: :article_create,\n        object: :article,\n        pre_hooks: []\n      },\n      %LetMe.Rule{\n        action: :update,\n        allow: [:own_resource],\n        deny: [],\n        name: :article_update,\n        object: :article,\n        pre_hooks: [:preload_groups]\n      }\n    ]","ref":"LetMe.Policy.html#c:list_rules/0-example"},{"type":"callback","title":"LetMe.Policy.list_rules/1","doc":"Same as `c:list_rules/0`, but takes a keyword list with filter options.\n\nSee `LetMe.filter_rules/2` for a list of available filter options.","ref":"LetMe.Policy.html#c:list_rules/1"},{"type":"macro","title":"LetMe.Policy.metadata/2","doc":"Assigns metadata to the action in the form of a key value pair.\n\nThe metadata can be accessed from the `LetMe.Rule` struct. You can use it\nto extend the functionality of the library.","ref":"LetMe.Policy.html#metadata/2"},{"type":"macro","title":"Example - LetMe.Policy.metadata/2","doc":"object :article do\n      action :create do\n        allow role: :writer\n\n        desc \"Allows a user to create a new article.\"\n        metadata :desc_ja, \"ユーザーが新しい記事を作成できるようにする\"\n      end\n    end\n\nThe `LetMe.Rule` struct returned by the introspection functions would look\nlike this:\n\n    %LetMe.Rule{\n      action: :create,\n      allow: [[role: :writer]],\n      deny: [],\n      description: \"Allows a user to create a new article.\",\n      name: :article_create,\n      object: :article,\n      pre_hooks: [],\n      metadata: [\n        desc_ja: \"ユーザーが新しい記事を作成できるようにする\"\n      ]\n    }\n\nIt is also possible to use the `metadata` macro multiple times.\n\n    object :article do\n      action :create do\n        allow role: :writer\n\n        desc \"Allows a user to create a new article.\"\n        metadata :desc_ja, \"ユーザーが新しい記事を作成できるようにする\"\n        metadata :desc_es, \"Permite al usuario crear un nuevo artículo.\"\n      end\n    end\n\nThis would result in:\n\n    %LetMe.Rule{\n      action: :create,\n      allow: [[role: :writer]],\n      deny: [],\n      description: \"Allows a user to create a new article.\",\n      name: :article_create,\n      object: :article,\n      pre_hooks: [],\n      metadata: [\n        desc_ja: \"ユーザーが新しい記事を作成できるようにする\",\n        desc_es: \"Permite al usuario crear un nuevo artículo.\"\n      ]\n    }","ref":"LetMe.Policy.html#metadata/2-example"},{"type":"macro","title":"LetMe.Policy.object/3","doc":"Defines an object on which actions can be performed.\n\nWithin the do-block, you can use the `action/2` macro to define the actions\nand checks.","ref":"LetMe.Policy.html#object/3"},{"type":"macro","title":"Examples - LetMe.Policy.object/3","doc":"object :article do\n      action :create do\n        allow role: :writer\n      end\n\n      action :delete do\n        allow role: :editor\n      end\n    end\n\nYou can optionally pass the schema module as the second argument. The schema\nmodule should implement the `LetMe.Schema` behaviour.\n\n    object :article, MyApp.Blog.Article do\n      action :create do\n        allow role: :writer\n      end\n    end\n\nAt the moment, this doesn't do much, except that you can find the schema\nmodule by passing the object name to `c:get_schema/1`, or find the object name\nby passing the schema module or struct to `c:get_object_name/1` now. Also,\nyou can now only pass the struct to`c:MyApp.Policy.filter_allowed_actions/3`,\nwithout explicitly passing the object name.","ref":"LetMe.Policy.html#object/3-examples"},{"type":"macro","title":"LetMe.Policy.pre_hooks/1","doc":"Registers one or multiple functions to run in order to hydrate the subject\nand/or object of the request.\n\nThis is useful if you need to enhance the data for multiple checks in the\nsame action by preloading associations, making external requests, or similar\nthings. The configured hook functions will be called once before running the\nchecks for an action.\n\nThe referenced functions must take the subject and object as arguments and\nreturn a 2-tuple with the updated subject and object.\n\nThe referenced functions may also take an optional third argument that are\nopts passed through the `authorize` functions. These opts are merged into\nany opts that are specified in the `pre_hook` definition.","ref":"LetMe.Policy.html#pre_hooks/1"},{"type":"macro","title":"Examples - LetMe.Policy.pre_hooks/1","doc":"Let's assume we defined these check and hook functions in our check module:\n\n    def MyApp.Policy.Checks do\n      # Checks\n\n      def min_age(%{age: age}, _, min_age), do: age >= min_age\n\n      # Hooks\n\n      def double_age(subject, object) do\n        new_subject = %{subject | age: subject.age * 2}\n        {new_subject, object}\n      end\n\n      def set_age(subject, object, age: age) do\n        new_subject = %{subject | age: age}\n        {new_subject, object}\n      end\n    end\n\nIf an atom is passed, LetMe will try to find the function in the check module.\n\n    object :article do\n      action :view do\n        pre_hooks :double_age\n        allow min_age: 50\n      end\n    end\n\nWith this in place, the following authorization request will evaluate to\n`true`:\n\n    MyApp.Policy.authorize!(:article_view, %{age: 25})\n    # => true\n\nIf your hooks are defined in a different module, you can also pass a\nmodule/function tuple. The pre-hook configuration above is equivalent to:\n\n    object :article do\n      action :view do\n        pre_hooks {MyApp.Policy.Checks, :double_age}\n        allow min_age: 50\n      end\n    end\n\nYou can also pass options to a hook by using an MFA tuple:\n\n    object :article do\n      action :view do\n        pre_hooks {MyApp.Policy.Checks, :set_age, age: 50}\n        allow min_age: 50\n      end\n    end\n\n    MyApp.Policy.authorize!(:article_view, %{age: 10})\n    # => true\n\nYou can achieve the same functionality dynamically using the opts on `authorize!`:\n\n    object :article do\n      action :view do\n        pre_hooks {MyApp.Policy.Checks, :set_age}\n        allow min_age: 50\n      end\n    end\n\n    MyApp.Policy.authorize!(:article_view, %{age: 10}, age: 50)\n    # => true\n\nAnd finally, you can also pass a list of hooks, which will be run in sequence:\n\n    alias MyApp.Policy.Checks\n\n    object :article do\n      action :view do\n        pre_hooks [{Checks, :set_age, 25}, :double_age]\n        allow min_age: 50\n      end\n    end\n\n    MyApp.Policy.authorize!(:article_view, %{age: 10})\n    # => true","ref":"LetMe.Policy.html#pre_hooks/1-examples"},{"type":"module","title":"LetMe.Rule","doc":"A struct for an authorization rule.","ref":"LetMe.Rule.html"},{"type":"type","title":"LetMe.Rule.check/0","doc":"A `check` references a function in the configured check module.\n\nCan be either one of:\n\n- A function name as an atom. The function must be a 2-arity function that\n  takes the subject (usually the current user) and the object as\n  arguments.\n- A tuple with the function name as an atom and a value of any type. The\n  function must be a 3-arity function that takes the subject, the object, and\n  the given value as arguments.","ref":"LetMe.Rule.html#t:check/0"},{"type":"type","title":"LetMe.Rule.hook/0","doc":"A hook can be registered to hydrate the subject and/or object before passing\nthem to the check functions.\n\n\"Hydration\" in this context means enriching or preparing the data by adding\nor transforming necessary information. For instance, you might fetch related\ndata from the database, calculate derived properties, or format the data in a\ncertain way.\n\nA hook can be one of the following:\n\n- The name of a function defined in the configured check module as an atom.\n- A `{module, function}` tuple.\n- A `{module, function, arguments}` tuple.\n\nIn either case, the function must take the subject as the first argument, the\nobject as the second argument, and return a tuple with the updated subject and\nobject. If an MFA tuple is passed, the given arguments are appended to the\ndefault arguments.","ref":"LetMe.Rule.html#t:hook/0"},{"type":"type","title":"LetMe.Rule.t/0","doc":"Struct for an authorization rule.\n\n- `action` - The action (verb) to be performed on the object, e.g. `:update`.\n- `allow` - A list of lists of checks to run to determine whether the action\n  is allowed. The outer list contains the alternatives (one for each `allow`\n  call; combined with `OR`). The inner lists are the checks for each `allow`\n  (combined with `AND`).\n- `description` - A human-readable description of the action.\n- `deny` - A list of lists of checks to run to determine whether the action is\n  explicitly denied. Same format as in `allow`. If any of these checks returns\n  `true`, the end result of the authorization request is immediately `false`,\n  even if any of the checks in the `allow` field would return `true`.\n- `name` - The name of the rule. Is always `{object}_{action}`.\n- `object` - The object that the action is performed on, e.g. `:article`.\n- `pre_hooks` - Functions to run in order to hydrate the subject and/or object\n  before running the allow and deny checks.\n- `metadata` - A list of relevant metadata useful for extending functionality.\n\nThe list entries in the outer list of the `allow` and `deny` fields are\ncombined with a logical `OR`. If one of the entries is a list of checks, those\nchecks are combined with a logical `AND`.","ref":"LetMe.Rule.html#t:t/0"},{"type":"type","title":"Examples - LetMe.Rule.t/0","doc":"- `[{role: :editor}, {role: :writer}]` - role is editor OR role is writer\n- `[[{role: :editor}], [{role: :writer}]]` - same as above\n- `[[{role: :editor}], [{role: :writer}, {:own_resource}]]` -\n   (role is editor OR (role is writer AND object is the user's own resource))","ref":"LetMe.Rule.html#t:t/0-examples"},{"type":"behaviour","title":"LetMe.Schema","doc":"Defines a behaviour with callbacks for scoping and redactions.\n\nUsing this module will define overridable default implementations for the\n`c:scope/3` and `c:redacted_fields/3` callbacks.","ref":"LetMe.Schema.html"},{"type":"behaviour","title":"Usage - LetMe.Schema","doc":"defmodule MyApp.Blog.Article do\n      use LetMe.Schema\n      import Ecto.Schema\n      alias MyApp.Accounts.User\n\n      @impl LetMe.Schema\n      def scope(q, user, opts \\\\ nil)\n      def scope(q, %User{role: :admin}, _), do: q\n      def scope(q, %User{}, _), do: where(q, published: true)\n\n      @impl LetMe.Schema\n      def redacted_fields(_, %User{role: :admin}, _), do: []\n      def redacted_fields(%__MODULE__{user_id: id}, %User{id: id}, _), do: []\n      def redacted_fields(_, %User{}, _), do: [:view_count]\n    end\n\n> #### `use LetMe.Schema` {: .info}\n>\n> When you `use LetMe.Schema`, the module will set `@behaviour LetMe.Schema`\n> and define default implementations for the functions `scope/3` and\n> `redacted_fields/3`. Both functions are overridable.","ref":"LetMe.Schema.html#module-usage"},{"type":"behaviour","title":"Scoping a query - LetMe.Schema","doc":"With the setup above, you can scope a blog article query depending on the\nuser.\n\n    defmodule MyApp.Blog do\n      import Ecto.Query\n\n      alias MyApp.Accounts.User\n      alias MyApp.Blog.Article\n\n      def list_articles(%User{} = current_user) do\n        Article\n        |> Article.scope(current_user)\n        |> Repo.all()\n      end\n\n      def get_article(id, %User{} = current_user) when is_integer(id) do\n        Article\n        |> where(id: id)\n        |> Article.scope(current_user)\n        |> Repo.one()\n      end\n    end","ref":"LetMe.Schema.html#module-scoping-a-query"},{"type":"behaviour","title":"Redacting fields - LetMe.Schema","doc":"After implementing the `c:redacted_fields/3` callback, you can hide fields\nfrom depending on the user by calling `LetMe.redact/3` on a struct or a list\nof structs.\n\n    def list_articles(%User{} = current_user) do\n      Article\n      |> Repo.all()\n      |> LetMe.redact(current_user)\n    end","ref":"LetMe.Schema.html#module-redacting-fields"},{"type":"callback","title":"LetMe.Schema.redacted_fields/3","doc":"Returns the fields that need to be removed from the given object for the given\nsubject.\n\nThis function can be used to hide certain fields depending on the current\nuser. See also `LetMe.redact/3` and `LetMe.reject_redacted_fields/3`.\n\n    defmodule MyApp.Blog.Article do\n      use LetMe.Schema\n      alias MyApp.Accounts.User\n\n      @impl LetMe.Schema\n      # hide view count unless the user is an admin or the article was written\n      # by the user\n      def redacted_fields(_, %User{role: :admin}, _), do: []\n      def redacted_fields(%__MODULE__{user_id: id}, %User{id: id}, _), do: []\n      def redacted_fields(_, %User{}, _), do: [:view_count]\n    end\n\nThe return value must be a list of fields to redact. You can also redact\nnested fields by passing the field names directly:\n\n    [:email, sibling: [:phone_number]]\n\nOr you can pass the module name of a nested struct, if that module also\nimplements `LetMe.Schema`:\n\n    [:email, sibling: MyApp.Relative]\n\nThe last argument can be used for any additional options.","ref":"LetMe.Schema.html#c:redacted_fields/3"},{"type":"callback","title":"LetMe.Schema.scope/3","doc":"Takes a queryable (usually an `Ecto.Queryable`) and a subject (usually the\ncurrent user) and returns an updated queryable.\n\nThis allows you to add `WHERE` clauses to a query depending on the user. For\nexample, you may want to add a `WHERE` clause to only return articles that are\npublished, unless the user is an admin. Or you may want to only return\nobjects that belong to the user.\n\n    defmodule MyApp.Blog.Article do\n      use Ecto.Schema\n      use LetMe.Schema\n\n      import Ecto.Schema\n      alias MyApp.Accounts.User\n\n      # Ecto schema and changeset\n\n      @impl LetMe.Schema\n      def scope(q, user, opts \\\\ nil)\n      def scope(q, %User{role: :admin}, _), do: q\n      def scope(q, %User{}, _), do: where(q, published: true)\n    end\n\nSince LetMe does not depend on Ecto and does not make any assumptions on the\nqueryable passed to the callback function, you are not constrained to use this\nmechanism for Ecto queries only. For example, you could use the function to\nadd filter parameters before passing them to a filter function or making an\nAPI call.\n\n    @impl LetMe.Schema\n    def scope(q, user, opts \\\\ nil)\n\n    def scope(query_params, %User{role: :admin}, _), do: query_params\n\n    def scope(query_params, %User{}, _) do\n      Keyword.put(query_params, :published, true)\n    end\n\nYou can use the third argument to pass any additional options.","ref":"LetMe.Schema.html#c:scope/3"},{"type":"type","title":"LetMe.Schema.redacted_fields/0","doc":"","ref":"LetMe.Schema.html#t:redacted_fields/0"},{"type":"exception","title":"LetMe.UnauthorizedError","doc":"Raised by `c:LetMe.Policy.authorize!/4` if a request is unauthorized.","ref":"LetMe.UnauthorizedError.html"},{"type":"extras","title":"LetMe","doc":"# LetMe\n\n[![CI](https://github.com/woylie/let_me/workflows/CI/badge.svg)](https://github.com/woylie/let_me/actions) [![Hex](https://img.shields.io/hexpm/v/let_me)](https://hex.pm/packages/let_me) [![Hex Docs](https://img.shields.io/badge/hex-docs-green)](https://hexdocs.pm/let_me/readme.html) [![Coverage Status](https://coveralls.io/repos/github/woylie/let_me/badge.svg)](https://coveralls.io/github/woylie/let_me)\n\nLetMe is a user-friendly authorization library for Elixir. Designed with a\nsimple and expressive Domain Specific Language (DSL), it provides an intuitive\nway to define and manage your authorization rules.\n\nThe strength of LetMe lies not only in its simplicity but also in its\nintrospection capabilities. It equips you with functions to answer important\nquestions about your application's authorization landscape, such as:\n\n- Which actions are defined in my application?\n- What are the conditions for a particular action?\n- Which actions are permissible for a user assigned a specific role?\n\nWith its intuitive DSL for rule definition coupled with introspection\ncapabilities, LetMe makes managing permissions in your application a breeze.","ref":"readme.html"},{"type":"extras","title":"Installation - LetMe","doc":"Add LetMe to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:let_me, \"~> 1.2.3\"}\n  ]\nend\n```\n\nInclude LetMe in your `.formatter.exs` file:\n\n```elixir\n[\n  import_deps: [:let_me]\n]\n```\n\nThis ensures that your LetMe authorization rules are formatted correctly when\nyou run `mix format`.\n\nNow, you're ready to start defining authorization rules with LetMe!","ref":"readme.html#installation"},{"type":"extras","title":"Policy module - LetMe","doc":"The Policy module sits at the heart of LetMe. It provides macros that allow you\nto define the authorization rules of your application. These rules are then\ncompiled into functions for both authorization checks and introspection.\n\nFor instance, here's how you might define a policy for a simple article CRUD interface:\n\n```elixir\ndefmodule MyApp.Policy do\n  use LetMe.Policy\n\n  object :article do\n    # Creating articles is allowed if the user role is `editor` or `writer`.\n    action :create do\n      allow role: :editor\n      allow role: :writer\n    end\n\n    # Viewing articles is always allowed, unless the user is banned.\n    action :read do\n      allow true\n      deny :banned\n    end\n\n    # Updating an article is allowed if (the user role is `editor`) OR\n    # (the user role is `writer` AND the article belongs to the user).\n    action :update do\n      allow role: :editor\n      allow [:own_resource, role: :writer]\n    end\n\n    # Deleting an article is allowed if the user is an editor.\n    action :delete do\n      allow role: :editor\n    end\n  end\nend\n```\n\nThe design of your policy modules—whether you have a single module for your\nentire application, one for each context, or some other arrangement—is\ncompletely up to you. LetMe offers the flexibility to organize your policy in\nthe way that best fits your application's needs.\n\nPlease note that while this example uses Role-Based Access Control (RBAC) for\nsimplicity, LetMe doesn't make any assumptions about your access control model.\nYou are completely free to define your authorization rules in any way you see\nfit.","ref":"readme.html#policy-module"},{"type":"extras","title":"Check module - LetMe","doc":"Authorization rules, generally speaking, are based on the subject (usually the\ncurrent user), the object on which the action is performed, and the action\nitself (the verb). LetMe doesn't enforce a particular authorization model or\ncheck implementation, instead allowing you to define what makes sense for your\napplication.\n\nThe checks passed to `LetMe.Policy.allow/1` reference functions in your own\ncheck module (by default `__MODULE__.Checks`, so in the given example, this\nwould be `MyApp.Policy.Checks`). Each function in your check module should\naccept the subject, the object, and optionally an extra argument. They must\nreturn a boolean value indicating the result of the check.\n\nFor the policy example provided earlier, a corresponding check module could look\nlike this:\n\n```elixir\ndefmodule MyApp.Policy.Checks do\n  alias MyApp.Accounts.User\n\n  @doc \"\"\"\n  Returns `true` if the `banned` flag is set on the user.\n  \"\"\"\n  def banned(%User{banned: banned}, _, _), do: banned\n\n  @doc \"\"\"\n  Checks whether the user ID of the object matches the ID of the current user.\n\n  Assumes that the object has a `:user_id` field.\n  \"\"\"\n  def own_resource(%User{id: id}, %{user_id: id}, _opts) when is_binary(id), do: true\n  def own_resource(_, _, _), do: false\n\n  @doc \"\"\"\n  Checks whether the user role matches the role passed as an option.","ref":"readme.html#check-module"},{"type":"extras","title":"Usage - LetMe","doc":"allow role: :editor\n\n  or\n\n      allow {:role, :editor}\n  \"\"\"\n  def role(%User{role: role}, _object, role), do: true\n  def role(_, _, _), do: false\nend\n```\n\nThis way, you can establish checks that are perfectly tailored to your\napplication's specific authorization requirements.","ref":"readme.html#usage"},{"type":"extras","title":"Callbacks - LetMe","doc":"When you incorporate `use LetMe.Policy` at the start of your policy module,\nLetMe generates a suite of useful functions for you:\n\n- Authorization functions: See `c:LetMe.Policy.authorize/4`,\n  `c:LetMe.Policy.authorize!/4` and `c:LetMe.Policy.authorize?/4`.\n- Introspection functions: See `c:LetMe.Policy.list_rules/0`,\n  `c:LetMe.Policy.list_rules/1`, `c:LetMe.Policy.get_rule/1` and others.\n\n#","ref":"readme.html#callbacks"},{"type":"extras","title":"Authorization - LetMe","doc":"You can employ the authorization functions wherever your application needs to\nmake authorization decisions. An ideal place to use these functions would be in\nyour context modules. Here's an example illustrating how you could incorporate\nauthorization into a blog's context module:\n\n```elixir\ndefmodule MyApp.Blog do\n  alias MyApp.Accounts.User\n  alias MyApp.Blog.Article\n  alias MyApp.Policy\n\n  def list_articles(%User{} = current_user) do\n    with :ok   {:error, :not_found}\n        article -> {:ok, article}\n      end\n    end\n  end\n\n  def create_article(params, %User{} = current_user) do\n    with :ok   Article.changeset(params)\n      |> Repo.insert()\n    end\n  end\n\n  def update_article(%Article{} = article, params, %User{} = current_user) do\n    with :ok   Article.changeset(params)\n      |> Repo.update()\n    end\n  end\n\n  def delete_article(%Article{} = article, %User{} = current_user) do\n    with :ok <- Policy.authorize(:article_delete, current_user, article) do\n      Repo.delete(article)\n    end\n  end\nend\n```\n\nIn this example, before performing any actions on the articles, we first ensure\nthe current user is authorized to perform the intended action. This makes our\napplication secure by making sure only authorized users can perform sensitive\noperations.\n\n##","ref":"readme.html#authorization"},{"type":"extras","title":"Typespecs - LetMe","doc":"LetMe automatically generates typespecs for the authorize functions in your\npolicy modules. This helps you to leverage Dialyzer's static type checking to\nensure valid actions are specified in your authorize calls. It's another way\nLetMe helps you write reliable, error-free code.\n\n#","ref":"readme.html#typespecs"},{"type":"extras","title":"Introspection - LetMe","doc":"LetMe equips you with introspection functions, enabling you to access the\ncomprehensive list of authorization rules. This can be beneficial, for instance,\nto render them on a documentation page:\n\n```elixir\niex> MyApp.Policy.list_rules()\n[\n  %LetMe.Rule{\n    action: :create,\n    allow: [\n      [role: :admin],\n      [role: :writer]\n    ],\n    deny: [],\n    description: nil,\n    name: :article_create,\n    object: :article,\n    pre_hooks: [],\n    metadata: []\n  },\n  # ...\n]\n```\n\nIf you wish to find a specific rule by its name, you can do so as well:\n\n```elixir\niex> MyApp.Policy.get_rule(:article_create)\n%LetMe.Rule{\n  action: :create,\n  allow: [\n    [role: :admin],\n    [role: :writer]\n  ],\n  name: :article_create,\n  object: :article,\n  # ...\n}\n```\n\nMoreover, you can list all actions associated with a particular role (or any\nother check):\n\n```elixir\niex> MyApp.Policy.list_rules(allow: {:role, :writer})\n[\n  %LetMe.Rule{\n    action: :create,\n    object: :article,\n    # ...\n  },\n  %LetMe.Rule{\n    action: :update,\n    object: :article,\n    # ...\n  }\n]\n```\n\nYou can also define metadata on an `action`. This feature can be used to extend\nthe library's functionality.\n\nFor example, imagine wanting to expose certain actions through your Absinthe\nGraphQL schema but needing to exclude others. You could achieve this by adding a\n`:gql_exclude` key to the metadata.\n\n```elixir\ndefmodule GraphqlPolicy do\n  use LetMe.Policy\n\n  object :user do\n    action :disable do\n      allow role: :admin\n      metadata :gql_exclude, true\n    end\n  end\nend\n```\n\n```elixir\niex> MyApp.Policy.get_rule(:user_disable)\n%LetMe.Rule{\n  action: :disable,\n  allow: [\n    [role: :admin]\n  ],\n  deny: [],\n  description: nil,\n  name: :user_disable,\n  object: :user,\n  pre_hooks: [],\n  metadata: [\n    gql_exclude: true\n  ]\n}\n```\n\nThis gives you the power to customize your authorization policies even further.","ref":"readme.html#introspection"},{"type":"extras","title":"Scoped queries - LetMe","doc":"There are situations where a user, despite having general access to a certain\nresource type, is only permitted to view a subset of the data. Consider a blog\nsystem: a user might be restricted to viewing only published articles, unless\nthey hold the role of a writer. Similarly, in a system where users are part of\nspecific companies, they might only be allowed to see users from their own\ncompany.\n\nTo tailor your queries based on the user type, implement the\n`c:LetMe.Schema.scope/3` callback of the LetMe.Schema behavior, typically within\nyour Ecto schema module.\n\n```elixir\ndefmodule MyApp.Blog.Article do\n  use Ecto.Schema\n  use LetMe.Schema\n\n  import Ecto.Query\n  alias MyApp.Accounts.User\n\n  # Ecto schema and changeset\n\n  @impl LetMe.Schema\n  def scope(q, user, opts \\\\ nil)\n  def scope(q, %User{role: :editor}, _), do: q\n  def scope(q, %User{role: :writer}, _), do: q\n  def scope(q, %User{}, _), do: where(q, published: true)\nend\n```\n\nIn this example, the Ecto query is modified to only return published articles,\nunless the user is an editor or writer. The third argument can be utilized for\nadditional options.\n\nWith this setup, your list and fetch functions can be updated as follows:\n\n```elixir\ndef list_articles(%User{} = current_user) do\n  with :ok   Article.scope(current_user)\n      |> Repo.all()\n\n    {:ok, articles}\n  end\nend\n\ndef fetch_article(id, %User{} = current_user) do\n  with :ok   where(id: ^id)\n      |> Article.scope(current_user)\n      |> Repo.one()\n\n    case result do\n      nil -> {:error, :not_found}\n      article -> {:ok, article}\n    end\n  end\nend\n```\n\nIf you've worked with [Bodyguard](https://hex.pm/packages/bodyguard) before,\nthis might look familiar. In Bodyguard,\nyou can find a `Bodyguard.scope/2` function that derives the Ecto schema module\nfrom the `Ecto.Queryable` and forwards the call to that module. In LetMe, you\nneed to call the `scope/2` function of your Ecto schema directly. The behaviour\nthen only serves to enforce this pattern.","ref":"readme.html#scoped-queries"},{"type":"extras","title":"Field redactions - LetMe","doc":"In certain scenarios, a user may be authorized to access a resource but should\nonly see a subset of its fields. For instance, one user might be able to see\nbasic details of another user, such as name and avatar, but shouldn't see\nsensitive information like email or phone number. One way to manage such cases\nwould be to conditionally show or hide specific information on the frontend.\nHowever, a cleaner solution is to have your context functions omit sensitive\nfields entirely.\n\nTo assist in these kinds of situations, the `LetMe.Schema` behaviour has another\ncallback: `c:LetMe.Schema.redacted_fields/3`.\n\n```elixir\ndefmodule MyApp.Accounts.User do\n  use Ecto.Schema\n  use LetMe.Schema\n\n  alias MyApp.Accounts.User\n\n  # Ecto schema and changeset\n\n  @impl LetMe.Schema\n  def redacted_fields(%User{}, %User{role: :admin}, _), do: []\n  def redacted_fields(%User{id: id}, %User{id: id}, _), do: []\n  def redacted_fields(%User{}, %User{}, _), do: [:email, :phone_number]\nend\n```\n\nThe `redacted_fields/2` function takes the object as the first argument, the\nsubject as the second argument, and an options argument. The function should\nreturn a list of fields to redact.\n\nIn the example above, all fields are visible if the user has an 'admin' role, or\nif the user being viewed (the object) is the same as the current user (the\nsubject). In other cases, the 'email' and 'phone_number' fields are hidden.\n\nThere are two strategies for handling field redactions:\n\n1. Modify the query to exclude the redacted fields.\n2. Redact the fields after retrieving the resource(s) from the database.\n\n#","ref":"readme.html#field-redactions"},{"type":"extras","title":"Modifying the query - LetMe","doc":"One approach to field redaction involves adjusting the database query to\nexclude redacted fields. Ecto's `__schema__/1` function can retrieve the\nnon-virtual schema fields from your Ecto module. From this list, you can reject\nany redacted fields and add a select clause that includes only the unredacted\nfields.\n\n```elixir\ndef list_users(%User{} = current_user) do\n  fields = User.__schema__(:fields)\n  filtered_fields = LetMe.reject_redacted_fields(fields, %User{}, current_user)\n\n  Article\n  |> select(^filtered_fields)\n  |> Repo.all()\nend\n```\n\nThis method has the advantage of preventing the transfer of redacted fields from\nthe database. However, it also comes with several drawbacks:\n\n1. Decisions about which fields to select cannot be made based on data in the\n   struct. For instance, with the `redacted_fields/2` function described\n   earlier, we can ensure that admins can see all fields, but we cannot\n   guarantee that users can view all fields in their own user account.\n2. All redacted fields will appear as `nil`, and you won't be able to\n   distinguish between fields that were redacted and fields that are simply\n   empty. This distinction might be necessary for display in the frontend.\n3. More complex select clauses may not be compatible with this syntax.\n\n#","ref":"readme.html#modifying-the-query"},{"type":"extras","title":"Redacting the query result - LetMe","doc":"To address the limitations of modifying the query, you can redact fields _after_\nretrieving the data from the database. This can be done using the\n`LetMe.redact/2` function.\n\n```elixir\ndef list_articles(%User{} = current_user) do\n  Article\n  |> Repo.all()\n  |> LetMe.redact(current_user)\nend\n```\n\nThe `redact` function can handle structs, lists of structs, and `nil` values.","ref":"readme.html#redacting-the-query-result"},{"type":"extras","title":"Why use this library? - LetMe","doc":"Consider using this library if:\n\n- You're seeking an easy-to-read DSL for authorization rules that offers the\n  flexibility to implement your authorization checks as desired.\n- You prefer to locate your authorization rules within your business layer,\n  thereby decoupling them from your interfaces.\n- You'd like to centralize your authorization rules in one place (or one per\n  context).\n- You want to generate a list of authorization rules.\n- You need to filter your authorization rules, e.g., to identify which actions a\n  certain user role can perform.\n- You're in need of a library that aids with query scopes and field redactions.\n- You prefer a library with zero dependencies.","ref":"readme.html#why-use-this-library"},{"type":"extras","title":"When not to use this library? - LetMe","doc":"This library might not be the best fit if:\n\n- You prefer to couple authorization checks with your interfaces.\n- You favor using plugs or middlewares for authorization checks and require\n  ready-made solutions (though you can create your own plugs and middlewares\n  around this library's functions).\n- You dislike DSLs and prefer to write functions directly (keep in mind, the DSL\n  only describes which checks to run and how to apply them; you'll still write\n  the actual checks as regular functions).\n- Introspection isn't a priority for you.\n- You need to provide details on why an authorization request fails. Checks in\n  LetMe currently return only a boolean value, meaning users receive a generic\n  error without knowing which exact check failed.","ref":"readme.html#when-not-to-use-this-library"},{"type":"extras","title":"Status - LetMe","doc":"This library is actively maintained. Given its zero dependencies and precisely\nscoped feature set, you may not see frequent updates. However, this is not an\nindication of stagnation but of stability. If you ever find something missing or\nencounter an issue, don't hesitate to open an issue – your feedback and\ncontributions are always welcome.","ref":"readme.html#status"},{"type":"extras","title":"Alternatives - LetMe","doc":"For comparison, consider exploring these Elixir libraries:\n\n- [Canada](https://hex.pm/packages/canada)\n- [Canary](https://hex.pm/packages/canary)\n- [Bodyguard](https://hex.pm/packages/bodyguard)\n- [Speakeasy](https://hex.pm/packages/speakeasy)\n\nThe article\n[Authorization for Phoenix Contexts](https://dockyard.com/blog/2017/08/01/authorization-for-phoenix-contexts)\nmay also be a helpful resource.","ref":"readme.html#alternatives"},{"type":"extras","title":"Changelog","doc":"# Changelog","ref":"changelog.html"},{"type":"extras","title":"Unreleased - Changelog","doc":"## [1.2.3] - 2023-11-11\n\n#","ref":"changelog.html#unreleased"},{"type":"extras","title":"Changed - Changelog","doc":"- Updated documentation.\n\n## [1.2.2] - 2023-06-28\n\n#","ref":"changelog.html#changed"},{"type":"extras","title":"Changed - Changelog","doc":"- You can now override the exception message used by\n  `c:LetMe.Policy.authorize!/4` (e.g.\n  `use LetMe.Policy, error_message: \"Not today, chap.\"`).\n\n## [1.2.1] - 2023-06-28\n\n#","ref":"changelog.html#changed"},{"type":"extras","title":"Changed - Changelog","doc":"- Define `action` type when you `use LetMe.Policy`.\n- Add type specifications for generated `authorize` functions.\n\n## [1.2.0] - 2023-06-19\n\n#","ref":"changelog.html#changed"},{"type":"extras","title":"Added - Changelog","doc":"- Added an optional `opts` argument to the authorize functions, so that\n  additional options can be passed to pre-hooks.\n- Updated `LetMe.filter_rules/2` to allow filtering by meta data.\n\n#","ref":"changelog.html#added"},{"type":"extras","title":"Changed - Changelog","doc":"- Pre-hook options are now expected to be passed as a keyword list.\n\n#","ref":"changelog.html#changed"},{"type":"extras","title":"Fixed - Changelog","doc":"- Fix deprecation warning about `Logger.warn/2` in Elixir 1.15.\n\n## [1.1.0] - 2023-05-08\n\n#","ref":"changelog.html#fixed"},{"type":"extras","title":"Added - Changelog","doc":"- Added a `metadata` macro to add metadata to actions. The metadata can be read\n  from the `LetMe.Rule` struct.\n\n## [1.0.3] - 2023-03-21\n\n#","ref":"changelog.html#added"},{"type":"extras","title":"Changed - Changelog","doc":"- Update `ex_doc` and other dev dependencies.\n\n## [1.0.2] - 2023-01-05\n\n#","ref":"changelog.html#changed"},{"type":"extras","title":"Added - Changelog","doc":"- Added a cheat sheet for rules and checks.\n\n#","ref":"changelog.html#added"},{"type":"extras","title":"Fixed - Changelog","doc":"- Fixed a code example for rule introspection in the readme.\n\n## [1.0.1] - 2022-11-06\n\n#","ref":"changelog.html#fixed"},{"type":"extras","title":"Changed - Changelog","doc":"- Use `Keyword.pop/3` with default value instead of `Keyword.pop!/2`, so that\n  you can pass options to `LetMe.redact/3` without passing the `redact_value`\n  option.\n\n## [1.0.0] - 2022-11-06\n\n#","ref":"changelog.html#changed"},{"type":"extras","title":"Added - Changelog","doc":"- Added `c:LetMe.Policy.filter_allowed_actions/3` and\n  `LetMe.filter_allowed_actions/4`.\n- Added `c:LetMe.Policy.get_object_name/1`.\n\n#","ref":"changelog.html#added"},{"type":"extras","title":"Changed - Changelog","doc":"- Renamed `c:LetMe.Policy.authorized?/3` to `c:LetMe.Policy.authorize?/3`,\n  because consistency is more important than grammar, maybe.\n- The `c:LetMe.Schema.scope/2` callback was removed in favour of\n  `c:LetMe.Schema.scope/3`. The `__using__` macro defined default\n  implementations for both functions that returned the given query unchanged, in\n  case you only needed the `redact` callback of the behaviour. In practice, this\n  made it all too easy to call the 2-arity version when only the 3-arity\n  version was defined, and vice versa, which would lead the query to not be\n  scoped. So in order to reduce the room for error at the cost of a minor\n  inconvenience, you will now always need to implement the 3-arity function,\n  even if you don't need the third argument.\n- Changed `c:LetMe.Schema.redacted_fields/2` to\n  `c:LetMe.Schema.redacted_fields/3` to allow passing additional options, and to\n  be consistent with `c:LetMe.Schema.scope/3`.\n\n## [0.2.0] - 2022-07-12\n\n#","ref":"changelog.html#changed"},{"type":"extras","title":"Changed - Changelog","doc":"- Added support for nested field redactions, either by explicitly listing the\n  fields or by referencing a module that also implements `LetMe.Schema`.\n\n#","ref":"changelog.html#changed"},{"type":"extras","title":"Fixed - Changelog","doc":"- `reject_redacted_fields/3` called `redact/2` callback with the wrong argument\n  order.\n\n## [0.1.0] - 2022-07-11\n\ninitial release","ref":"changelog.html#fixed"},{"type":"extras","title":"Rules and Checks","doc":"# Rules and Checks","ref":"rules.html"},{"type":"extras","title":"Minimal example - Rules and Checks","doc":"{: .col-2}\n\n#","ref":"rules.html#minimal-example"},{"type":"extras","title":"Policy definition - Rules and Checks","doc":"##","ref":"rules.html#policy-definition"},{"type":"extras","title":"Policy module - Rules and Checks","doc":"```elixir\ndefmodule MyApp.Policy do\n  use LetMe.Policy\n\n  object :article do\n    action :create do\n      allow role: :writer\n    end\n  end\nend\n```\n\n##","ref":"rules.html#policy-module"},{"type":"extras","title":"Check module - Rules and Checks","doc":"```elixir\ndefmodule MyApp.Policy.Checks do\n  def role(%MyApp.User{role: role}, _object, role), do: true\n  def role(_, _, _), do: false\nend\n```\n\n#","ref":"rules.html#check-module"},{"type":"extras","title":"Context module - Rules and Checks","doc":"```elixir\ndefmodule MyApp.Blog do\n  alias MyApp.Blog.Article\n  alias MyApp.Policy\n\n  def create_article(params, %MyApp.User{} = current_user) do\n    with :ok   Article.changeset(params)\n      |> Repo.insert()\n    end\n  end\nend\n```","ref":"rules.html#context-module"},{"type":"extras","title":"Check examples - Rules and Checks","doc":"{: .col-2}\n\n#","ref":"rules.html#check-examples"},{"type":"extras","title":"Check without options - Rules and Checks","doc":"##","ref":"rules.html#check-without-options"},{"type":"extras","title":"Policy module - Rules and Checks","doc":"```elixir\nobject :article do\n  action :create do\n    allow :writer\n  end\nend\n```\n\n##","ref":"rules.html#policy-module"},{"type":"extras","title":"Check implementation - Rules and Checks","doc":"```elixir\ndef writer(%MyApp.User{role: :writer}, _object, _opts), do: true\ndef writer(_, _, _), do: false\n```\n\n#","ref":"rules.html#check-implementation"},{"type":"extras","title":"Check with options - Rules and Checks","doc":"##","ref":"rules.html#check-with-options"},{"type":"extras","title":"Policy module - Rules and Checks","doc":"```elixir\nobject :article do\n  action :delete do\n    allow trust_level: 50\n  end\nend\n```\n\n##","ref":"rules.html#policy-module"},{"type":"extras","title":"Check implementation - Rules and Checks","doc":"```elixir\ndef trust_level(%MyApp.User{trust_level: actual_level}, _, required_level)\n  when actual_level >= required_level,\n  do: true\n\ndef trust_level(_, _, _), do: false\n```\n\n#","ref":"rules.html#check-implementation"},{"type":"extras","title":"Check that depends on the object - Rules and Checks","doc":"##","ref":"rules.html#check-that-depends-on-the-object"},{"type":"extras","title":"Policy module - Rules and Checks","doc":"```elixir\nobject :article do\n  action :update do\n    allow :own_resource\n  end\nend\n```\n\n##","ref":"rules.html#policy-module"},{"type":"extras","title":"Check implementation - Rules and Checks","doc":"```elixir\ndef own_resource(%MyApp.User{id: user_id}, %{user_id: user_id}, _), do: true\ndef own_resource(_, _, _), do: false\n```","ref":"rules.html#check-implementation"},{"type":"extras","title":"Rule examples - Rules and Checks","doc":"{: .col-2}\n\n#","ref":"rules.html#rule-examples"},{"type":"extras","title":"Multiple actions with the same rules - Rules and Checks","doc":"```elixir\nobject :article do\n  action [:create, :update, :delete] do\n    allow :admin\n  end\nend\n```\n\n#","ref":"rules.html#multiple-actions-with-the-same-rules"},{"type":"extras","title":"Combine checks with `AND` - Rules and Checks","doc":"```elixir\naction :create do\n  allow [:two_fa_enabled, role: :writer]\nend\n```\n\n#","ref":"rules.html#combine-checks-with-and"},{"type":"extras","title":"Combine checks with `OR` - Rules and Checks","doc":"```elixir\naction :create do\n  allow role: :admin\n  allow role: :writer\nend\n```\n\n#","ref":"rules.html#combine-checks-with-or"},{"type":"extras","title":"Conditionally allow with exception - Rules and Checks","doc":"```elixir\naction :create do\n  allow :is_admin\n  deny :is_suspended\nend\n```\n\n#","ref":"rules.html#conditionally-allow-with-exception"},{"type":"extras","title":"Always allow - Rules and Checks","doc":"```elixir\naction :read do\n  allow true\nend\n```\n\n#","ref":"rules.html#always-allow"},{"type":"extras","title":"Always deny - Rules and Checks","doc":"```elixir\naction :read do\n  deny true\nend\n```\n\n#","ref":"rules.html#always-deny"},{"type":"extras","title":"Always allow with exception - Rules and Checks","doc":"```elixir\naction :read do\n  allow true\n  deny :user_is_suspended\nend\n```\n\n#","ref":"rules.html#always-allow-with-exception"},{"type":"extras","title":"Add description - Rules and Checks","doc":"```elixir\naction :create do\n  desc \"allows a user to create a new article\"\n  allow role: :writer\nend\n```","ref":"rules.html#add-description"},{"type":"extras","title":"Pre-hooks - Rules and Checks","doc":"{: .col-2}\n\n#","ref":"rules.html#pre-hooks"},{"type":"extras","title":"Without options - Rules and Checks","doc":"##","ref":"rules.html#without-options"},{"type":"extras","title":"Policy module - Rules and Checks","doc":"```elixir\nobject :article do\n  action :create do\n    pre_hooks :preload_roles\n\n    allow role: :admin\n    allow role: :editor\n    allow role: :writer\n  end\nend\n```\n\n##","ref":"rules.html#policy-module"},{"type":"extras","title":"Check module - Rules and Checks","doc":"```elixir\ndef role(%MyApp.User{roles: roles}, _object, role) do\n  Enum.any?(roles, & &1.id == role)\nend\n\ndef role(_, _, _), do: false\n\ndef preload_roles(subject, object) do\n  {MyApp.Repo.preload(subject, [:roles]), object}\nend\n```\n\n#","ref":"rules.html#check-module"},{"type":"extras","title":"With options - Rules and Checks","doc":"##","ref":"rules.html#with-options"},{"type":"extras","title":"Policy module - Rules and Checks","doc":"```elixir\nobject :article do\n  action :create do\n    pre_hooks {:preload_roles, force: true}\n\n    allow role: :admin\n    allow role: :editor\n    allow role: :writer\n  end\nend\n```\n\n##","ref":"rules.html#policy-module"},{"type":"extras","title":"Check module - Rules and Checks","doc":"```elixir\ndef role(%MyApp.User{roles: roles}, _object, role) do\n  Enum.any?(roles, & &1.id == role)\nend\n\ndef role(_, _, _), do: false\n\ndef preload_roles(subject, object, opts) do\n  {MyApp.Repo.preload(subject, [:roles], opts), object}\nend\n```\n\n#","ref":"rules.html#check-module"},{"type":"extras","title":"From a different module - Rules and Checks","doc":"```elixir\nobject :article do\n  action :create do\n    pre_hooks {MyApp.Policy.Prehooks, :preload_roles, force: true}\n\n    allow role: :admin\n    allow role: :editor\n    allow role: :writer\n  end\nend\n```\n\n#","ref":"rules.html#from-a-different-module"},{"type":"extras","title":"Multiple pre-hooks - Rules and Checks","doc":"```elixir\nobject :article do\n  action :create do\n    pre_hooks [:preload_roles, :role_list_to_role_id_list]\n\n    allow role: :admin\n    allow role: :editor\n    allow role: :writer\n  end\nend\n```","ref":"rules.html#multiple-pre-hooks"},{"type":"extras","title":"Metadata - Rules and Checks","doc":"{: .col-2}\n\n```elixir\nobject :article do\n  action :create do\n    allow role: :admin\n    allow role: :editor\n    allow role: :writer\n    \n    metadata :gql_exclude, true\n    metadata :desc_ja, \"ユーザーが新しい記事を作成できるようにする\"\n  end\nend\n```","ref":"rules.html#metadata"}],"content_type":"text/markdown"}